using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Events;
using System;
using SimpleJSON;
using System.Collections.Generic;

using MacGruber;
using Obi;
using Battlehub.RTSaveLoad;

namespace ChenRong
{
    public class CanvasButton : MVRScript
    {
        #region PluginInfo
        public string pluginAuthor = "ChenRong";
        public string pluginName = "CanvasButton";
        public string pluginVersion = "3.2";
        public string pluginDate = "2024-01-18";
        public string pluginDescription = @"Canvas Button Plugin";
        #endregion

        private JSONStorableString ui_name;

        private JSONStorableString new_name;

        private JSONStorableFloat spacing = null;
        private JSONStorableFloat[] size = new JSONStorableFloat[2];
        private JSONStorableFloat[] position = new JSONStorableFloat[3];

        private JSONStorableStringChooser chooser;
        private List<object> ui_objects = new List<object>();

        private bool active = true;
        private Canvas canvas = null;
        private JSONStorableBool is_panelopen;
        private JSONStorableBool is_horizontally;
        private JSONStorableBool is_living3D;

        private Color text_color = Color.white;
        private Color button_color = new Color(1f, 0f, 0f, 0.2f);

        private JSONStorableFloat button_alpha = null;

        private JSONStorableFloat counts = null;

        private List<Widget> buttons = new List<Widget>();

        float fov = 40;
        float last_fov = 40;

        public override void Init()
        {
            SimpleTriggerHandler.LoadAssets();

            generate_variable();
            generate_uiplugin();

            this.ui_name.valNoCallback = "Action 1";
            this.new_name.valNoCallback = "NewAction";
        }

        public void Start()
        {
            restore_canvas();
            restore_uiplugin();

            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
        }

        private void Update()
        {
            var flash = false;
            foreach (var v in this.buttons)
                if (v.update())
                    flash = true;

            this.fov = SuperController.singleton.monitorCameraFOV;
            if (this.fov != this.last_fov)
            {
                this.last_fov = this.fov;
                flash = true;
            }

            if (flash)
            {
                restore_canvas();
                restore_uiplugin();
            }
        }

        private void FixedUpdate() { }

        void OnDestroy()
        {
            foreach (var v in this.buttons)
                v.delete();

            clear_canvas();
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
        }

        private void set_active(bool active)
        {
            this.active = active;
            foreach (var v in this.buttons)
                v.active(this.active);
        }

        private string create_name(string name)
        {
            var str = System.Text.RegularExpressions.Regex.Replace(name, @"[^0-9]+", "");
            if (str.Length == 0)
                return name + " 1";
            return name.Substring(0, name.Length - str.Length) + (int.Parse(str) + 1).ToString();
        }

        private void generate_button()
        {
            var name = this.ui_name.val;
            foreach (var b in this.buttons)
                if (b.alive(name))
                {
                    SuperController.LogError($"Button is alive, rename please");
                    return;
                }

            this.buttons.Add(new ButtonWidget(this, name, this.text_color, this.button_color));
            this.counts.val += 1;

            if (is_panelopen.val)
                this.buttons[this.buttons.Count - 1].trigger_panel();
            sync_trigger_trace();

            restore_canvas();
            restore_uiplugin();

            this.ui_name.valNoCallback = create_name(name);
            this.chooser.val = this.buttons[this.buttons.Count - 1].name;
        }

        private void generate_slider()
        {
            var name = this.ui_name.val;
            foreach (var b in this.buttons)
                if (b.alive(name))
                {
                    SuperController.LogError($"Button is alive, rename please");
                    return;
                }

            this.buttons.Add(new SliderWidget(this, name, this.text_color, this.button_color));
            this.counts.val += 1;

            if (is_panelopen.val)
                this.buttons[this.buttons.Count - 1].trigger_panel();
            sync_trigger_trace();

            restore_canvas();
            restore_uiplugin();

            this.ui_name.valNoCallback = create_name(name);
            this.chooser.val = this.buttons[this.buttons.Count - 1].name;
        }

        private void rename_button()
        {
            var name = this.chooser.val;
            foreach (var v in this.buttons)
                if (v.alive(this.new_name.val))
                {
                    SuperController.LogMessage(
                        $"Name " + this.new_name.val + " is illegal, please rename it"
                    );
                    return;
                }

            foreach (var v in this.buttons)
                if (v.alive(name))
                    v.rename(this.new_name.val);

            this.chooser.val = this.new_name.val;

            this.new_name.val = create_name(this.new_name.val);

            restore_canvas();
            restore_uiplugin();
        }

        private void destroy_button()
        {
            var label = this.chooser.val;
            foreach (var v in this.buttons)
                if (v.alive(label))
                {
                    v.delete();
                    this.buttons.Remove(v);
                    break;
                }

            restore_canvas();
            restore_uiplugin();
        }

        private void destroy_buttons()
        {
            foreach (var v in this.buttons)
                v.delete();
            this.buttons.Clear();

            this.ui_name.valNoCallback = "Action 1";
            this.new_name.valNoCallback = "NewAction";

            restore_canvas();
            restore_uiplugin();
        }

        private void copy_button()
        {
            foreach (var v in this.buttons)
                if (v.alive(this.new_name.val))
                {
                    SuperController.LogMessage(
                        $"Name " + this.new_name.val + " is illegal, please rename it"
                    );
                    return;
                }

            int i = 0;
            foreach (var v in this.buttons)
            {
                if (v.alive(this.chooser.val))
                {
                    break;
                }
                i++;
            }
            if (i == this.buttons.Count)
            {
                SuperController.LogMessage($"Can't find the button!");
            }

            var jc = this.buttons[i].copy(this.new_name.val);
            if (this.buttons[i].id == "ButtonWidget")
                this.buttons.Add(new ButtonWidget(this));
            else if (this.buttons[i].id == "SliderWidget")
                this.buttons.Add(new SliderWidget(this));
            else
            {
                SuperController.LogError($"Error: Can't find parameter class");
                return;
            }
            this.buttons[this.buttons.Count - 1].by(jc);
            counts.val += 1;
            this.new_name.val = create_name(this.new_name.val);

            restore_canvas();
            restore_uiplugin();

            // SuperController.LogMessage($"but " + but.name);
            // SuperController.LogMessage($"json button " + jc.ToString());
        }

        private void sord_buttons(bool is_up)
        {
            int i = 0;
            int target = 0;

            foreach (var v in this.buttons)
            {
                if (v.alive(this.chooser.val))
                {
                    break;
                }
                i++;
            }
            if (i > this.buttons.Count)
            {
                SuperController.LogMessage($"Can't find the button!");
                return;
            }

            if (is_up)
            {
                if (i == 0)
                    return;
                target = i - 1;
            }
            else
            {
                if (i == this.buttons.Count - 1)
                    return;
                target = i + 1;
            }

            var jc = this.buttons[i].copy(this.buttons[i].name);
            Widget b = null;
            if (this.buttons[i].id == "ButtonWidget")
                b = new ButtonWidget(this);
            else if (this.buttons[i].id == "SliderWidget")
                b = new SliderWidget(this);
            else
            {
                SuperController.LogError($"Error: Can't find parameter class");
                return;
            }
            b.by(jc);
            this.buttons.Remove(this.buttons[i]);
            this.buttons.Insert(target, b);

            restore_canvas();
            restore_uiplugin();
        }

        private void generate_canvas()
        {
            GameObject canvasObject = new GameObject();
            canvas = canvasObject.AddComponent<Canvas>();
            canvas.renderMode = RenderMode.WorldSpace;
            canvas.pixelPerfect = false;
            SuperController.singleton.AddCanvas(canvas);

            if (!is_living3D.val)
                canvas.transform.SetParent(Camera.main.transform, false); // optional

            CanvasScaler cs = canvasObject.AddComponent<CanvasScaler>();
            cs.scaleFactor = 80.0f;
            cs.dynamicPixelsPerUnit = 1f;

            GraphicRaycaster gr = canvasObject.AddComponent<GraphicRaycaster>(); // click enable

            float scale = 0.001f * SuperController.singleton.worldScale;
            canvas.transform.localScale = new Vector3(scale, scale, scale);
            canvas.transform.localPosition = new Vector3(0, 0, 0);
            canvas.transform.localEulerAngles = new Vector3(0, 0, 0); // look at camera

            if (is_living3D.val)
                canvas.transform.Translate(-2, 0, 0);
            else
            {
                if (SuperController.singleton.isOVR || SuperController.singleton.isOpenVR)
                    canvas.transform.Translate(-2f, 0.2f, 1f);
                else
                    // canvas.transform.Translate(-1.31f, 0.625f, 2f);
                    canvas.transform.Translate(-1.31f, 0.625f, 2f / ((fov / 40 - 1) * 1.2f + 1));
            }
        }

        private void clear_canvas()
        {
            if (canvas == null)
                return;

            if (SuperController.singleton != null)
                SuperController.singleton.RemoveCanvas(canvas);

            canvas.transform.SetParent(null, false);

            if (canvas.gameObject != null)
                GameObject.Destroy(canvas.gameObject);
        }

        private void restore_canvas()
        {
            clear_canvas();
            generate_canvas();

            for (int i = 0; i < this.buttons.Count; i++)
                this.buttons[i].add(
                    this.canvas,
                    this.position,
                    this.spacing,
                    i,
                    is_horizontally.val
                );
        }

        private void sync_trigger_trace()
        {
            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
        }

        private void generate_variable()
        {
            this.ui_name = new JSONStorableString("Name", "");
            this.new_name = new JSONStorableString("Name", "");

            this.counts = new JSONStorableFloat("counts", 0, 0, 100);
            RegisterFloat(this.counts);

            this.position[0] = new JSONStorableFloat(
                "Position (x)",
                1f,
                (float v) =>
                {
                    restore_canvas();
                },
                0f,
                1f,
                false,
                true
            );
            this.position[1] = new JSONStorableFloat(
                "Position (y)",
                0f,
                (float v) =>
                {
                    restore_canvas();
                },
                0f,
                1f,
                false,
                true
            );
            this.position[2] = new JSONStorableFloat(
                "Position (z)",
                0f,
                (float v) =>
                {
                    restore_canvas();
                },
                -1f,
                1f,
                false,
                true
            );

            this.spacing = new JSONStorableFloat(
                "Spacing",
                0.05f,
                (float v) =>
                {
                    restore_canvas();
                },
                0f,
                0.2f,
                false,
                true
            );

            this.size[0] = new JSONStorableFloat(
                "Width",
                40,
                (float v) =>
                {
                    foreach (var b in this.buttons)
                    {
                        b.geometry(this.size[0].val, this.size[1].val);
                    }
                    restore_canvas();
                },
                20,
                80,
                false,
                true
            );

            this.size[1] = new JSONStorableFloat(
                "Height",
                40,
                (float v) =>
                {
                    foreach (var b in this.buttons)
                    {
                        b.geometry(this.size[0].val, this.size[1].val);
                    }
                    restore_canvas();
                },
                30,
                120,
                false,
                true
            );

            button_alpha = new JSONStorableFloat(
                "Alpha",
                button_color.a,
                (float a) =>
                {
                    // foreach (var v in this.buttons)
                    //     v.color_alpha(a);
                    this.button_color.a = a;
                    foreach (var v in this.buttons)
                        v.color(this.button_color, false);
                },
                0f,
                1f,
                false,
                true
            );

            is_horizontally = new JSONStorableBool(
                "Horizontally Button",
                false,
                (bool v) =>
                {
                    restore_canvas();
                }
            );
            RegisterBool(is_horizontally);

            is_living3D = new JSONStorableBool(
                "Living3D Button",
                false,
                (bool v) =>
                {
                    restore_canvas();
                }
            );
            RegisterBool(is_living3D);

            RegisterFloat(this.button_alpha);
            RegisterFloat(this.spacing);
            foreach (var v in this.position)
                RegisterFloat(v);
            foreach (var v in this.size)
                RegisterFloat(v);

            // this.text_color = new JSONStorableColor("text_color", new HSVColor());
            // this.button_color = new JSONStorableColor("button_color", new HSVColor());
            // // this.button_color.SetColor(new Color(0.8392f, 0.8392f, 0.8392f));

            // RegisterColor(this.text_color);
            // RegisterColor(this.button_color);

            this.chooser = new JSONStorableStringChooser(
                "Selected State",
                new List<string>(),
                string.Empty,
                "Selected Button",
                (string v) =>
                {
                    restore_uiplugin();
                }
            );

            RegisterAction(
                new JSONStorableAction(
                    "Toggle",
                    () =>
                    {
                        set_active(!active);
                    }
                )
            );
            RegisterAction(
                new JSONStorableAction(
                    "Show",
                    () =>
                    {
                        set_active(true);
                    }
                )
            );
            RegisterAction(
                new JSONStorableAction(
                    "Hide",
                    () =>
                    {
                        set_active(false);
                    }
                )
            );
        }

        private void generate_uiplugin()
        {
            Utils.OnInitUI(CreateUIElement);

            Utils.SetupInfoText(
                this,
                "<color=#000><size=35><b>CanvasButton</b></size></color>\n"
                    + "<color=#333>"
                    + "Add Button to 2D UI interface\n"
                    + "</color>",
                120.0f,
                true
            );

            Utils.SetupInfoOneLine(this, "<size=30><b>Button Config</b></size>", false);
            Utils.SetupTextInput(this, "Name", this.ui_name, false);

            Utils.SetupSpacer(this, 3, false);
            JSONStorableBool is_text_color = Utils.SetupToggle(
                this,
                "Button Text Color",
                false,
                false
            );

            JSONStorableColor color = Utils.SetupColor(this, "Button Color", Color.red, false);
            color.setCallbackFunction += (float h, float s, float v) =>
            {
                foreach (var b in this.buttons)
                {
                    var c = HSVColorPicker.HSVToRGB(h, s, v);
                    c.a = this.button_alpha.val;
                    b.color(c, is_text_color.val);
                }
                if (is_text_color.val)
                {
                    this.text_color = HSVColorPicker.HSVToRGB(h, s, v);
                }
                else
                {
                    this.button_color = HSVColorPicker.HSVToRGB(h, s, v);
                    this.button_color.a = this.button_alpha.val;
                }
            };

            Utils.SetupSpacer(this, 3, false);

            Utils.SetupTwinButton(
                this,
                "Generate Button",
                () =>
                {
                    generate_button();
                },
                "Generate Slider",
                () =>
                {
                    generate_slider();
                },
                false
            );
            Utils.SetupTwinButton(
                this,
                "Remove All",
                () =>
                {
                    destroy_buttons();
                },
                "Toggle All",
                () =>
                {
                    set_active(!active);
                },
                false
            );

            Utils.SetupSpacer(this, 5, false);

            // Utils.SetupInfoOneLine(this, "<size=30><b>Button Quick Trigger</b></size>", false);

            Utils.SetupInfoOneLine(this, "<size=30><b>Button Information</b></size>", true);
            CreateFilterablePopup(this.chooser, true);

            Utils.SetupTwinButton(
                this,
                "Upward",
                () =>
                {
                    sord_buttons(true);
                },
                "Downward",
                () =>
                {
                    sord_buttons(false);
                },
                true
            );
            Utils.SetupButton(
                this,
                "Remove Button",
                () =>
                {
                    destroy_button();
                },
                true
            );

            Utils.SetupTextInput(this, "New Name", this.new_name, true);
            Utils.SetupTwinButton(
                this,
                "Rename",
                () =>
                {
                    rename_button();
                },
                "Copy",
                () =>
                {
                    copy_button();
                },
                true
            );

            // Utils.SetupSpacer(this, 1, true);

            Utils.SetupInfoOneLine(this, "<size=30><b>Button Extra Config</b></size>", false);
            is_panelopen = Utils.SetupToggle(this, "Open Panel When Creating", true, false);
            CreateToggle(is_horizontally, false);
            CreateToggle(is_living3D, false);

            CreateSlider(this.spacing, false);
            CreateSlider(this.button_alpha, false);
            foreach (var v in this.position)
                CreateSlider(v, false);
            foreach (var v in this.size)
                CreateSlider(v, false);
        }

        private void restore_uiplugin()
        {
            var choices = new List<string>();

            foreach (var v in this.buttons)
                choices.Add(v.name);

            this.chooser.choices = null;
            this.chooser.choices = choices;
            this.chooser.displayChoices = choices;

            var update_chooser = true;
            foreach (var v in this.buttons)
                if (v.alive(this.chooser.val))
                    update_chooser = false;

            if (this.buttons.Count == 0)
                this.chooser.val = "";
            else if (this.chooser.val == "" || update_chooser)
                this.chooser.val = choices[0];

            Utils.RemoveUIElements(this, ui_objects);

            foreach (var v in this.buttons)
                if (v.alive(this.chooser.val))
                    v.config_base(ui_objects);

            ui_objects.Add(Utils.SetupSpacer(this, 10, true));
            ui_objects.Add(
                Utils.SetupInfoOneLine(this, "<size=30><b>Button Quick Trigger</b></size>", true)
            );

            foreach (var b in this.buttons)
                b.config_trigger(ui_objects);
        }

        private void OnAtomRename(string oldid, string newid)
        {
            foreach (var v in this.buttons)
                v.sync();
        }

        public override JSONClass GetJSON(
            bool includePhysical = true,
            bool includeAppearance = true,
            bool forceStore = false
        )
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if (includePhysical || forceStore)
            {
                needsStore = true;

                for (int i = 0; i < this.buttons.Count; i++)
                    jc["classid" + i.ToString()] = this.buttons[i].id;

                for (int i = 0; i < this.buttons.Count; i++)
                    this.buttons[i].to_json(jc, i);
            }
            // SuperController.LogMessage($"json " + jc.ToString());
            return jc;
        }

        public override void RestoreFromJSON(
            JSONClass jc,
            bool restorePhysical = true,
            bool restoreAppearance = true,
            JSONArray presetAtoms = null,
            bool setMissingToDefault = true
        )
        {
            base.RestoreFromJSON(
                jc,
                restorePhysical,
                restoreAppearance,
                presetAtoms,
                setMissingToDefault
            );

            for (int i = 0; i < this.counts.val; i++)
            {
                string s = jc["classid" + i.ToString()];
                // SuperController.LogError($"restre " + s + " no: " + i.ToString());

                if (s == "ButtonWidget")
                    this.buttons.Add(new ButtonWidget(this));
                else if (s == "SliderWidget")
                    this.buttons.Add(new SliderWidget(this));
                else
                {
                    SuperController.LogError($"Error: Can't find classid!");
                    continue;
                }
                this.buttons[i].from_json(jc, i, base.subScenePrefix, base.mergeRestore);
            }

            this.text_color = this.buttons[0].text;
            this.button_color = this.buttons[0].background;
            this.ui_name.valNoCallback = create_name(this.buttons[this.buttons.Count - 1].name);
        }
    }
}
