using System.Linq;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;

using MacGruber;
using System.Collections.Generic;
using System.Collections;
using UnityEngine.UI;

namespace ChenRong
{
    public class Widget
    {
        public string id;
        public string name;

        public float width = 40;
        public float height = 40;
        public Color text = Color.white;
        public Color background = Color.red;

        public bool flash = false;

        public bool is_show = true;

        public MVRScript script;

        public UIDynamicButton vr_button;

        public void configure_transform(Transform t, float width, float height)
        {
            t.transform.position = Vector3.zero;
            RectTransform rt = t.GetComponent<RectTransform>();
            rt.anchoredPosition = new Vector2(width / 2, height / 2);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
        }

        public virtual void trigger_panel(bool is_toggle = false) { }

        public virtual void geometry(float width, float height)
        {
            this.width = width;
            this.height = height;
        }

        public virtual void rename(string name)
        {
            this.name = name;
        }

        public virtual bool alive(string name)
        {
            if (name == this.name)
                return true;
            return false;
        }

        public virtual bool update()
        {
            return this.flash;
        }

        public virtual void delete() { }

        public virtual void active(bool v) { }

        public virtual void sync() { }

        public virtual void color(Color c, bool is_text = false)
        {
            if (is_text)
            {
                c.a = 1;
                this.text = c;
            }
            else
                this.background = c;
            this.flash = true;
        }

        public virtual JSONClass copy(string name)
        {
            var record = this.name;
            JSONClass jc = new JSONClass();

            this.rename(name);
            this.to_json(jc, 0);
            this.rename(record);

            return jc;
        }

        public virtual void by(JSONClass jc)
        {
            this.from_json(jc, 0, "", false);
        }

        public virtual void add(
            Canvas canvas,
            JSONStorableFloat[] position,
            JSONStorableFloat spacing,
            int index,
            bool direction
        ) { }

        public virtual void config_base(List<object> objects) { }

        public virtual void config_trigger(List<object> objects) { }

        public virtual void to_json(JSONClass json, int no) { }

        public virtual void from_json(
            JSONClass json,
            int no,
            string subScenePrefix,
            bool isMerge
        ) { }
    }

    public class TextWidget : Widget
    {
        // public string id = "TextWidget";
    }

    public class SliderWidget : Widget
    {
        private JSONStorableFloat value = null;
        private UIDynamicSlider slider = null;

        private FloatTrigger handle;

        private const string enum_handle = "Normal";

        public SliderWidget(MVRScript script)
        {
            this.id = "SliderWidget";
            this.script = script;
        }

        public SliderWidget(MVRScript script, string name, Color text, Color background)
        {
            this.id = "SliderWidget";
            this.script = script;
            this.name = name;
            this.text = text;
            this.background = background;
            this.handle = new FloatTrigger(script, this.name);

            rename(name);
        }

        private void event_rename(CustomTrigger t, string e, int no)
        {
            t.Name = this.name + "[" + this.id + ":" + e + no.ToString() + "]";
        }

        private UIDynamicSlider create_slider(
            Canvas canvas,
            UnityAction<float> callback,
            float width = 100,
            float height = 80
        )
        {
            Transform slider = GameObject.Instantiate<Transform>(
                this.script.manager.configurableSliderPrefab
            );

            configure_transform(slider, width, height * 1.9f);
            slider.SetParent(canvas.transform, false); // ParentToCanvas

            slider.transform.Find("Panel").GetComponent<Image>().enabled = false;

            Text[] textComponents = slider.GetComponentsInChildren<Text>(true);
            foreach (Text textComponent in textComponents)
                if (
                    textComponent.name == "Text"
                    && textComponent.transform.parent.name == "ValueInputField"
                )
                    textComponent.enabled = false;

            UIDynamicSlider ui_slider = slider.GetComponent<UIDynamicSlider>();

            ui_slider.Configure(this.name, 0, 1, value.val, true, "F2", false, false);
            ui_slider.defaultButtonEnabled = false;
            ui_slider.label = this.name;
            ui_slider.slider.image.color = new Color(0, 0, 0, 0);
            ui_slider.slider.fillRect.GetComponent<Image>().color = this.background;

            ui_slider.slider.onValueChanged.AddListener(callback);
            ui_slider.slider.onValueChanged.AddListener(
                (float v) =>
                {
                    this.value.val = v;
                }
            );

            return ui_slider;
        }

        public override void sync()
        {
            this.handle.SyncAtomNames();
        }

        public override void active(bool v)
        {
            this.is_show = v;
            this.slider.gameObject.SetActive(v);
        }

        public override void rename(string name)
        {
            this.name = name;
            // this.button.label = this.name;
            this.handle.Name = this.name;
            // event_rename(this.handle, enum_handle, 0);
            if (this.value != null)
                this.script.DeregisterFloat(this.value);
            this.value = new JSONStorableFloat(
                this.name,
                0,
                (float v) =>
                {
                    if (this.slider == null)
                        return;
                    this.slider.slider.value = v;
                },
                0,
                1,
                false,
                true
            );
            this.script.RegisterFloat(this.value);
        }

        private void trigger(float v)
        {
            this.handle.Trigger(v);
        }

        public override void config_base(List<object> objects) { }

        public override void config_trigger(List<object> objects)
        {
            objects.Add(
                Utils.SetupButton(
                    this.script,
                    "[ " + this.name + " ]   Trigger",
                    () =>
                    {
                        this.handle.OpenPanel();
                    },
                    true
                )
            );
        }

        public override void trigger_panel(bool is_toggle = false)
        {
            this.handle.OpenPanel();
        }

        public override void delete()
        {
            if (this.slider == null)
                return;
            Object.Destroy(this.slider.gameObject);
        }

        public override void add(
            Canvas canvas,
            JSONStorableFloat[] position,
            JSONStorableFloat spacing,
            int index,
            bool direction
        )
        {
            float row = 0;
            float column = 0;
            float position_x = position[0].val * 2.58f;
            float position_y = -position[1].val * 1.4f;
            float position_z = -position[2].val;

            if (direction)
                column = -index * spacing.val * 3;
            else
                row = -index * spacing.val;

            delete();

            this.slider = create_slider(
                canvas,
                (float v) =>
                {
                    trigger(v);
                },
                width,
                height
            );
            this.slider.transform.Translate(
                position_x + column,
                position_y + row + -0.03f,
                position_z,
                Space.Self
            );

            this.slider.labelText.color = this.text;
            this.slider.labelText.fontSize = 20;

            this.flash = false;

            active(this.is_show);
        }

        public override void to_json(JSONClass json, int no)
        {
            var id = this.id + no.ToString();
            json[id]["name"] = name;

            json[id]["text"]["r"].AsFloat = this.text.r;
            json[id]["text"]["g"].AsFloat = this.text.g;
            json[id]["text"]["b"].AsFloat = this.text.b;
            json[id]["text"]["a"].AsFloat = this.text.a;
            json[id]["geometry"]["w"].AsFloat = this.width;
            json[id]["geometry"]["h"].AsFloat = this.height;
            json[id]["background"]["r"].AsFloat = this.background.r;
            json[id]["background"]["g"].AsFloat = this.background.g;
            json[id]["background"]["b"].AsFloat = this.background.b;
            json[id]["background"]["a"].AsFloat = this.background.a;
            json[id]["is_show"].AsBool = this.is_show;

            event_rename(this.handle, enum_handle, no);
            json[this.handle.Name] = this.handle.GetJSON();
            this.handle.Name = this.name;

            // SuperController.LogMessage($"json" + json.ToString());
        }

        public override void from_json(JSONClass json, int no, string subScenePrefix, bool isMerge)
        {
            var id = this.id + no.ToString();
            this.name = json[id]["name"];

            this.text.r = json[id]["text"]["r"].AsFloat;
            this.text.g = json[id]["text"]["g"].AsFloat;
            this.text.b = json[id]["text"]["b"].AsFloat;
            this.text.a = json[id]["text"]["a"].AsFloat;
            this.width = json[id]["geometry"]["w"].AsFloat;
            this.height = json[id]["geometry"]["h"].AsFloat;
            this.background.r = json[id]["background"]["r"].AsFloat;
            this.background.g = json[id]["background"]["g"].AsFloat;
            this.background.b = json[id]["background"]["b"].AsFloat;
            this.background.a = json[id]["background"]["a"].AsFloat;
            this.is_show = json[id]["is_show"].AsBool;

            this.handle = new FloatTrigger(this.script, this.name);
            event_rename(this.handle, enum_handle, no);
            this.handle.RestoreFromJSON(json, subScenePrefix, isMerge, true);
            this.handle.Name = this.name;
            rename(this.name);
        }
    }

    public class ButtonWidget : Widget
    {
        // public string id = "ButtonWidget";

        private EventTrigger handle;
        private UIDynamicButton button;
        private JSONStorableAction action;

        private bool toggle_enable = false;
        private bool toggle_status = false;
        private string toggle_name = null;
        private EventTrigger toggle_handle = null;

        private bool delay_enable = false;
        private JSONStorableFloat delay_time = null;
        private EventTrigger delay_n_handle = null;
        private EventTrigger delay_t_handle = null;

        public ButtonWidget(MVRScript script)
        {
            this.id = "ButtonWidget";
            this.script = script;

            this.delay_time = new JSONStorableFloat("Delay", 0.3f, 0f, 10f, false, true);
        }

        public ButtonWidget(MVRScript script, string name, Color text, Color background)
        {
            this.id = "ButtonWidget";
            this.script = script;
            this.name = name;
            this.text = text;
            this.background = background;
            this.handle = new EventTrigger(script, this.name);

            this.delay_time = new JSONStorableFloat("Delay", 0.3f, 0f, 10f, false, true);
            rename(this.name);
        }

        public override void rename(string name)
        {
            this.name = name;
            this.handle.Name = this.name;

            if (this.action != null)
                this.script.DeregisterAction(this.action);
            this.action = new JSONStorableAction(
                this.name,
                () =>
                {
                    trigger();
                }
            );
            this.script.RegisterAction(this.action);
        }

        public override void active(bool v)
        {
            this.is_show = v;
            this.button.gameObject.SetActive(v);
        }

        public override bool update()
        {
            this.handle.Update();
            if (toggle_enable)
                this.toggle_handle.Update();
            if (delay_enable)
            {
                this.delay_n_handle.Update();
                if (toggle_enable)
                    this.delay_t_handle.Update();
            }
            return flash;
        }

        private IEnumerator n_trigger()
        {
            yield return new WaitForSeconds(this.delay_time.val);
            this.delay_n_handle.Trigger();
        }

        private IEnumerator t_trigger()
        {
            yield return new WaitForSeconds(this.delay_time.val);
            this.delay_t_handle.Trigger();
        }

        public void trigger()
        {
            if (!this.toggle_enable)
            {
                this.handle.Trigger();
                if (this.delay_enable)
                    script.StartCoroutine(n_trigger());
                return;
            }

            if (!toggle_status)
            {
                this.handle.Trigger();
                if (this.delay_enable)
                    script.StartCoroutine(n_trigger());
            }
            else
            {
                this.toggle_handle.Trigger();
                if (this.delay_enable)
                    script.StartCoroutine(t_trigger());
            }

            toggle_status = !toggle_status;
            flash = true;
        }

        public override void trigger_panel(bool is_toggle = false)
        {
            var handle = this.handle;
            if (is_toggle)
                handle = this.toggle_handle;

            handle.OpenPanel();
        }

        public override void sync()
        {
            this.handle.SyncAtomNames();
            if (toggle_enable)
                this.toggle_handle.SyncAtomNames();
            if (delay_enable)
            {
                this.delay_n_handle.SyncAtomNames();
                if (toggle_enable)
                    this.delay_t_handle.SyncAtomNames();
            }
        }

        public override void add(
            Canvas canvas,
            JSONStorableFloat[] position,
            JSONStorableFloat spacing,
            int index,
            bool direction
        )
        {
            float row = 0;
            float column = 0;
            float position_x = position[0].val * 2.58f;
            float position_y = -position[1].val * 1.4f;
            float position_z = -position[2].val;

            if (direction)
                column = -index * spacing.val * 3;
            else
                row = -index * spacing.val;

            delete();

            this.button = create_button(
                canvas,
                () =>
                {
                    trigger();
                },
                width,
                height
            );
            this.button.transform.Translate(
                position_x + column,
                position_y + row,
                position_z,
                Space.Self
            );

            if (!toggle_status)
                this.button.label = this.name;
            else
                this.button.label = this.toggle_name;

            this.button.textColor = this.text;
            this.button.buttonColor = this.background;

            this.flash = false;

            active(this.is_show);

            this.vr_button = this.button;
        }

        public override void delete()
        {
            if (this.button == null)
                return;
            Object.Destroy(this.button.gameObject);
        }

        public override void config_base(List<object> objects)
        {
            if (!this.toggle_enable)
            {
                JSONStorableString name = new JSONStorableString("Name", "~ " + this.name);
                objects.Add(Utils.SetupTextInput(this.script, "Toggle", name, true));
                objects.Add(
                    Utils.SetupButton(
                        this.script,
                        "Generate Toggle",
                        () =>
                        {
                            this.toggle_enable = true;
                            this.toggle_name = name.val;
                            this.toggle_handle = new EventTrigger(
                                this.script,
                                this.toggle_name + "T"
                            );
                            this.flash = true;

                            this.toggle_handle.OpenPanel();
                        },
                        true
                    )
                );
            }
            else
            {
                JSONStorableString name = new JSONStorableString("Name", "NewToggle");
                objects.Add(Utils.SetupTextInput(this.script, "New Toggle", name, true));
                objects.Add(
                    Utils.SetupTwinButton(
                        this.script,
                        "Rename",
                        () =>
                        {
                            this.toggle_name = name.val;
                            this.toggle_handle.Name = this.toggle_name + "T";
                            this.flash = true;
                        },
                        "Remove",
                        () =>
                        {
                            this.toggle_enable = false;
                            this.toggle_status = false;
                            this.toggle_name = null;
                            this.toggle_handle = null;

                            this.flash = true;
                        },
                        true
                    )
                );
            }

            if (!delay_enable)
                objects.Add(
                    Utils.SetupButton(
                        this.script,
                        "Generate Delay Trigger",
                        () =>
                        {
                            this.delay_n_handle = new EventTrigger(this.script, this.name + "=");
                            this.delay_t_handle = new EventTrigger(this.script, this.name + "T=");
                            this.delay_enable = true;
                            this.flash = true;
                        },
                        true
                    )
                );
            else
            {
                objects.Add(script.CreateSlider(this.delay_time, true));

                objects.Add(
                    Utils.SetupButton(
                        this.script,
                        "Remove Delay Trigger",
                        () =>
                        {
                            this.delay_n_handle = null;
                            this.delay_t_handle = null;
                            this.delay_enable = false;
                            this.flash = true;
                        },
                        true
                    )
                );
            }
        }

        public override void config_trigger(List<object> objects)
        {
            if (!this.toggle_enable)
            {
                if (!this.delay_enable)
                    objects.Add(
                        Utils.SetupButton(
                            this.script,
                            "[ " + this.name + " ]   Trigger",
                            () =>
                            {
                                this.trigger_panel();
                            },
                            true
                        )
                    );
                else
                    objects.Add(
                        Utils.SetupTwinButton(
                            this.script,
                            "[ " + this.name + " ]",
                            () =>
                            {
                                this.trigger_panel();
                            },
                            "[ " + this.name + " = ]",
                            () =>
                            {
                                this.delay_n_handle.OpenPanel();
                            },
                            true
                        )
                    );
            }
            else
            {
                objects.Add(
                    Utils.SetupTwinButton(
                        this.script,
                        "[ " + this.name + " ]",
                        () =>
                        {
                            this.trigger_panel();
                        },
                        "[ " + this.toggle_name + " ]",
                        () =>
                        {
                            this.trigger_panel(true);
                        },
                        true
                    )
                );
                if (this.delay_enable)
                {
                    objects.Add(
                        Utils.SetupTwinButton(
                            this.script,
                            "[ " + this.name + " = ]",
                            () =>
                            {
                                this.delay_n_handle.OpenPanel();
                            },
                            "[ " + this.toggle_name + " = ]",
                            () =>
                            {
                                this.delay_t_handle.OpenPanel();
                            },
                            true
                        )
                    );
                }
            }
        }

        private UIDynamicButton create_button(
            Canvas canvas,
            UnityAction callback,
            float width = 100,
            float height = 80
        )
        {
            Transform button = GameObject.Instantiate<Transform>(
                this.script.manager.configurableButtonPrefab
            );
            configure_transform(button, width, height);
            button.SetParent(canvas.transform, false); // ParentToCanvas

            UIDynamicButton uiButton = button.GetComponent<UIDynamicButton>();
            uiButton.label = this.name;
            uiButton.buttonText.fontSize = 20;
            uiButton.button.onClick.AddListener(callback);

            return uiButton;
        }

        public override void to_json(JSONClass json, int no)
        {
            var id = this.id + no.ToString();
            json[id]["name"] = name;

            json[id]["text"]["r"].AsFloat = this.text.r;
            json[id]["text"]["g"].AsFloat = this.text.g;
            json[id]["text"]["b"].AsFloat = this.text.b;
            json[id]["text"]["a"].AsFloat = this.text.a;
            json[id]["geometry"]["w"].AsFloat = this.width;
            json[id]["geometry"]["h"].AsFloat = this.height;
            json[id]["background"]["r"].AsFloat = this.background.r;
            json[id]["background"]["g"].AsFloat = this.background.g;
            json[id]["background"]["b"].AsFloat = this.background.b;
            json[id]["background"]["a"].AsFloat = this.background.a;
            json[id]["toggle_enable"].AsBool = this.toggle_enable;
            json[id]["delay_enable"].AsBool = this.delay_enable;
            json[id]["is_show"].AsBool = this.is_show;

            this.handle.Name = this.name + no.ToString();
            json[this.handle.Name] = this.handle.GetJSON();
            this.handle.Name = this.name;

            // SuperController.LogMessage($"handle " + this.handle.Name);
            // SuperController.LogMessage($"toggle handle " + this.toggle_handle.Name);

            if (this.toggle_enable)
            {
                json[id]["toggle_name"] = this.toggle_name;
                json[id]["toggle_status"].AsBool = this.toggle_status;
                this.toggle_handle.Name = this.toggle_name + "T" + no.ToString();
                json[this.toggle_handle.Name] = this.toggle_handle.GetJSON();
                this.toggle_handle.Name = this.toggle_name;
            }
            if (this.delay_enable)
            {
                // SuperController.LogMessage($"delay_n_handle " + this.delay_n_handle.Name);
                json[id]["delay_time"].AsFloat = this.delay_time.val;
                this.delay_n_handle.Name = this.name + "=" + no.ToString();
                json[this.delay_n_handle.Name] = this.delay_n_handle.GetJSON();
                this.delay_n_handle.Name = this.name + "=";

                if (this.toggle_enable)
                {
                    // SuperController.LogMessage($"thandle " + this.delay_t_handle.Name);
                    this.delay_t_handle.Name = this.name + "T=" + no.ToString();
                    json[this.delay_t_handle.Name] = this.delay_t_handle.GetJSON();
                    this.delay_t_handle.Name = this.name + "T=";
                }
            }
        }

        public override void from_json(JSONClass json, int no, string subScenePrefix, bool isMerge)
        {
            var id = this.id + no.ToString();
            this.name = json[id]["name"];

            this.text.r = json[id]["text"]["r"].AsFloat;
            this.text.g = json[id]["text"]["g"].AsFloat;
            this.text.b = json[id]["text"]["b"].AsFloat;
            this.text.a = json[id]["text"]["a"].AsFloat;
            this.width = json[id]["geometry"]["w"].AsFloat;
            this.height = json[id]["geometry"]["h"].AsFloat;
            this.background.r = json[id]["background"]["r"].AsFloat;
            this.background.g = json[id]["background"]["g"].AsFloat;
            this.background.b = json[id]["background"]["b"].AsFloat;
            this.background.a = json[id]["background"]["a"].AsFloat;
            this.toggle_enable = json[id]["toggle_enable"].AsBool;
            this.delay_enable = json[id]["delay_enable"].AsBool;
            this.is_show = json[id]["is_show"].AsBool;

            this.handle = new EventTrigger(this.script, this.name + no.ToString());
            this.handle.RestoreFromJSON(json, subScenePrefix, isMerge, true);

            rename(this.name);
            // active(this.is_show);

            // SuperController.LogMessage($"subScenePrefix " + subScenePrefix);
            // SuperController.LogMessage($"isMerge " + isMerge.ToString());
            // SuperController.LogMessage($"restore handle " + this.handle.Name);
            if (this.toggle_enable)
            {
                this.toggle_name = json[id]["toggle_name"];
                this.toggle_status = json[id]["toggle_status"].AsBool;
                this.toggle_handle = new EventTrigger(
                    this.script,
                    this.toggle_name + "T" + no.ToString()
                );
                this.toggle_handle.RestoreFromJSON(json, subScenePrefix, isMerge, true);
                this.toggle_handle.Name = this.toggle_name;
                // SuperController.LogMessage($"restore toggle_handle " + this.toggle_handle.Name);
            }

            if (this.delay_enable)
            {
                this.delay_time.val = json[id]["delay_time"].AsFloat;
                this.delay_n_handle = new EventTrigger(
                    this.script,
                    this.name + "=" + no.ToString()
                );
                this.delay_n_handle.RestoreFromJSON(json, subScenePrefix, isMerge, true);
                this.delay_n_handle.Name = this.name + "=";

                if (this.toggle_enable)
                {
                    // SuperController.LogMessage($"thandle " + this.delay_t_handle.Name);

                    this.delay_t_handle = new EventTrigger(
                        this.script,
                        this.name + "T=" + no.ToString()
                    );
                    this.delay_t_handle.RestoreFromJSON(json, subScenePrefix, isMerge, true);
                    this.delay_t_handle.Name = this.name + "T=";
                }
            }
        }
    }
}
